# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from abc import ABC
from abc import abstractmethod
import argparse
import functools
from itertools import takewhile
import os
import os.path
from typing import Any, Callable, Dict, List, Optional, Tuple

import psutil

from austin.config import AustinConfiguration

try:
    _cached_property = functools.cached_property  # type: ignore[attr-defined]
except AttributeError:

    def _cached_property(f: Callable[..., Any]) -> property:  # type: ignore[no-redef]
        return property(functools.lru_cache(maxsize=1)(f))


SemVer = Tuple[int, int, int]


def _to_semver(version: Optional[str]) -> SemVer:
    if version is None:
        return (0, 0, 0)

    try:
        return (  # type: ignore[return-value]
            tuple(
                int(_)
                for _ in "".join(
                    list(
                        takewhile(
                            lambda _: _.isdigit() or _ == ".", version.replace("?", "0")
                        )
                    )
                ).split(".")
            )
            + (0, 0, 0)
        )[:3]
    except ValueError:
        raise ValueError("Invalid semantic version")


class AustinError(Exception):
    """Basic Austin Error."""

    pass


class AustinTerminated(AustinError):
    """Austin termination exception.

    Thrown when Austin is terminated with a call to ``terminate``.
    """

    pass


class BaseAustin(ABC):
    """Base Austin class.

    Defines the general API that abstract Austin as an external process.
    Subclasses should implement the :func:`start` method and either define the
    :func:`on_sample_received` method or pass it via the constructor.
    Additionally, the :func:`on_ready` and the :func:`on_terminate` methods can
    be overridden or passed via the constructor to catch the corresponding
    events. Austin is considered to be ready when the first sample is received;
    it is considered to have terminated when the process has terminated
    gracefully.

    If an error was encountered, the :class:`AustinError` exception is thrown.
    """

    BINARY = "austin"
    BINARY_VERSION = (3, 0, 0)

    def __init__(
        self,
        sample_callback: Callable[[str], None] = None,
        ready_callback: Callable[[psutil.Process, psutil.Process, str], None] = None,
        terminate_callback: Callable[[Dict[str, str]], None] = None,
    ) -> None:
        """The ``BaseAustin`` constructor.

        The callbacks can be passed as arguments, if they are not
        defined/overridden in the definition of a sub-class.
        """
        self._proc: psutil.Process = None
        self._child_proc: psutil.Process = None
        self._cmd_line: Optional[str] = None
        self._running: bool = False
        self._meta: Dict[str, str] = {}

        try:
            self._sample_callback = sample_callback or self.on_sample_received  # type: ignore[attr-defined]
        except AttributeError:
            raise AustinError("No sample callback given or implemented.")

        self._terminate_callback = terminate_callback or self.on_terminate
        self._ready_callback = ready_callback or self.on_ready

        self._version: Optional[str] = None
        self._python_version: Optional[str] = None

    def _get_process_info(
        self, args: argparse.Namespace, austin_pid: int
    ) -> Tuple[psutil.Process, psutil.Process, str]:
        try:
            self._proc = psutil.Process(austin_pid)
        except psutil.NoSuchProcess:
            raise AustinError("Cannot find Austin process.")

        if not args.pid:  # Austin is forking
            try:
                self._child_proc = self._proc.children()[0]
                if self._child_proc is None:
                    raise IndexError
            except IndexError:
                raise AustinError("Cannot find Austin child process.")
        else:  # Austin is attaching
            try:
                self._child_proc = psutil.Process(args.pid)
            except psutil.NoSuchProcess:
                raise AustinError(
                    f"Cannot attach to process with invalid PID {args.pid}."
                )

        self._cmd_line = " ".join(self._child_proc.cmdline())

        self._running = True

        return self._proc, self._child_proc, self._cmd_line

    @abstractmethod
    def start(self, args: List[str] = None) -> Any:
        """Start Austin.

        Every subclass should implement this method and ensure that it spawns
        a new Austin process.
        """
        ...

    def get_process(self) -> psutil.Process:
        """Get the underlying Austin process.

        Return an instance of :class:`psuitl.Process` that can be used to
        control the underlying Austin process at the OS level.
        """
        return self._proc

    def get_command_line(self) -> Optional[str]:
        """Get the inferred command line.

        Return the command line of the (main) process that is being profiled
        by Austin.
        """
        return self._cmd_line

    def is_running(self) -> bool:
        """Determine whether Austin is running."""
        return self._running

    def terminate(self, wait: bool = False) -> None:
        """Terminate Austin.

        Stop the underlying Austin process by sending a termination signal.
        """
        if not self._proc:
            raise AustinError("Austin has not been started yet")

        try:
            self._proc.terminate()
            if wait:
                self._proc.wait()
        except psutil.NoSuchProcess:
            raise AustinError("Austin has already terminated")

        self._running = False
        self._proc = None
        self._child_proc = None

    def get_child_process(self) -> psutil.Process:
        """Get the child process.

        Return an instalce of :class:`psutil.Process` representing the (main)
        process being profiled by Austin at the OS level.
        """
        return self._child_proc

    def submit_sample(self, data: bytes) -> None:
        """Submit a sample to the sample callback.

        This method takes care of converting the raw binary data retrieved from
        Austin into a Python string.
        """
        try:
            sample = data.decode()
        except UnicodeDecodeError:
            try:
                sample = data.decode("ascii")
            except UnicodeDecodeError:
                return

        self._sample_callback(sample.rstrip())

    def check_exit(self, rcode: int, stderr: Optional[str]) -> None:
        """Check Austin exit status."""
        if rcode:
            if rcode in (-15, 15):
                raise AustinTerminated()
            raise AustinError(f"({rcode}) {stderr}")

    def check_version(self) -> None:
        """Check for the minimum Austin binary version."""
        austin_version = self.version
        if austin_version is None:
            raise AustinError("Cannot determine Austin version")
        if austin_version < self.BINARY_VERSION:
            raise AustinError(
                f"Incompatible Austin version (got {austin_version}, expected >= {self.BINARY_VERSION})"
            )

    # ---- Default callbacks ----

    def on_terminate(self, stats: Dict[str, str]) -> Any:
        """Terminate event callback.

        Implement to be notified when Austin has terminated gracefully. The
        callback accepts an argument that will receive the global statistics.
        """
        pass

    def on_ready(
        self,
        process: psutil.Process,
        child_process: psutil.Process,
        command_line: str,
        data: Any = None,
    ) -> Any:
        """Ready event callback.

        Implement to get notified when Austin has successfully started or
        attached the Python process to profile and the first sample has been
        produced. This callback receives the Austin process and it's (main)
        profiled process as instances of :class:`psutil.Process`, along with
        the command line of the latter.
        """
        pass

    # ---- Properties ----

    @_cached_property
    def binary_path(self) -> str:
        """Discover the path of the Austin binary.

        Lookup order is:
        - current working directory
        - ``AUSTINPATH`` variable
        - ``~/.austinrc`` file
        - ``PATH`` variable.
        """
        binary_name = self.BINARY + (os.name == "nt" and ".exe" or "")

        # Try CWD
        binary_path = os.path.join(os.getcwd(), binary_name)
        if os.path.isfile(binary_path):
            return binary_path

        # Try with AUSTINPATH variable
        austin_path = os.environ.get("AUSTINPATH", None)
        if austin_path:
            binary_path = os.path.join(os.path.expanduser(austin_path), binary_name)
            if os.path.isfile(binary_path):
                return binary_path

        # Try with .austinrc file
        if AustinConfiguration().binary is not None:
            binary_path = os.path.expanduser(AustinConfiguration().binary)
            if os.path.isfile(binary_path):
                return binary_path

        # Try from PATH
        return self.BINARY

    @property
    def version(self) -> SemVer:
        """Austin version."""
        return _to_semver(self._meta.get("austin"))

    @property
    def python_version(self) -> SemVer:
        """The version of the detected Python interpreter."""
        return _to_semver(self._meta.get("python"))



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import sys
from typing import Dict, List

from austin import AustinError
from austin import BaseAustin
from austin.cli import AustinArgumentParser


class SimpleAustin(BaseAustin):
    """Simple implementation of Austin.

    This is the simplest way to start Austin from Python if you do not need to
    carry out other operation in parallel. Calling :func:`start` returns only
    once Austin has terminated.

    Example::

        class EchoSimpleAustin(SimpleAustin):
            def on_ready(self, process, child_process, command_line):
                print(f"Austin PID: {process.pid}")
                print(f"Python PID: {child_process.pid}")
                print(f"Command Line: {command_line}")

            def on_sample_received(self, line):
                print(line)

            def on_terminate(self, data):
                print(data)

        try:
            austin = EchoSimpleAustin()
            austin.start(["-i", "10000"], ["python3", "myscript.py"])
        except KeyboardInterrupt:
            pass
    """

    def _read_meta(self) -> Dict[str, str]:
        assert self.proc.stdout

        meta = {}

        while True:
            line = self.proc.stdout.readline().decode().rstrip()
            if not (line and line.startswith("# ")):
                break
            key, _, value = line[2:].partition(": ")
            meta[key] = value

        self._meta.update(meta)

        return meta

    def start(self, args: List[str] = None) -> None:
        """Start the Austin process."""
        try:
            self.proc = subprocess.Popen(
                [self.binary_path] + ["-P"] + (args or sys.argv[1:]),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except FileNotFoundError:
            raise AustinError("Austin executable not found.")

        if not self.proc.stdout:
            raise AustinError("Standard output stream is unexpectedly missing")
        if not self.proc.stderr:
            raise AustinError("Standard error stream is unexpectedly missing")

        try:
            if not self._read_meta():
                raise AustinError("Austin did not start properly")

            self.check_version()

            self._ready_callback(
                *self._get_process_info(
                    AustinArgumentParser().parse_args(args), self.proc.pid
                )
            )

            while self.is_running():
                data = self.proc.stdout.readline().rstrip()
                if not data:
                    break

                self.submit_sample(data)

            self._terminate_callback(self._read_meta())
            try:
                stderr = self.proc.communicate(timeout=1)[1].decode().rstrip()
            except subprocess.TimeoutExpired:
                stderr = ""
            self.check_exit(self.proc.wait(), stderr)

        except Exception:
            self.proc.terminate()
            self.proc.wait()
            raise

        finally:
            self._running = False



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, BinaryIO, Dict, List, Tuple, Union

from austin.format import Mode
from austin.format.pprof.profile_pb2 import Profile
from austin.format.pprof.profile_pb2 import Sample
from austin.stats import Frame
from austin.stats import Sample as AustinSample

PROCESS_ID_LABEL = "Process ID"
THREAD_ID_LABEL = "Thread ID"

CPU_TYPE = "CPU time"
WALL_TYPE = "Wall time"
TIME_UNIT = "μs"

MEMORY_ALLOC_TYPE = "Allocated memory"
MEMORY_DEALLOC_TYPE = "Deallocated memory"
MEMORY_UNIT = "B"


class PProf:
    """PProf generator class.

    Used to convert Austin collapsed format into the pprof protobuf format.
    """

    def __init__(self, mode: Union[Mode, str]) -> None:
        self._string_table: Dict[str, int] = {}
        self._location_map: Dict[Frame, int] = {}
        self._function_map: Dict[Tuple[str, str], int] = {}

        # Create the protobuf Profile message
        self.profile = Profile()
        self.profile.string_table.append("")

        # Add sample types according to the mode
        if isinstance(mode, str):
            mode = Mode.from_metadata(mode)

        if mode == Mode.MEMORY:
            self._add_memory_sample_types()

        elif mode == Mode.CPU:
            self._add_time_sample_type(CPU_TYPE)

        elif mode == Mode.WALL:
            self._add_time_sample_type(WALL_TYPE)

        elif mode == Mode.FULL:
            self._add_time_sample_type(CPU_TYPE)
            self._add_time_sample_type(WALL_TYPE)
            self._add_memory_sample_types()

        self.mode = mode

    def _add_sample_type(self, type: str, unit: str) -> None:
        _ = self.profile.sample_type.add()
        _.type = self.get_string(type)
        _.unit = self.get_string(unit)

    def _add_time_sample_type(self, type: str) -> None:
        self._add_sample_type(type, TIME_UNIT)

    def _add_memory_sample_types(self) -> None:
        self._add_sample_type(MEMORY_ALLOC_TYPE, MEMORY_UNIT)
        self._add_sample_type(MEMORY_DEALLOC_TYPE, MEMORY_UNIT)

    def get_string(self, string: str) -> int:
        """Get the string table index for the given string."""
        try:
            return self._string_table[string]
        except KeyError:
            index = len(self.profile.string_table)
            self._string_table[string] = index
            self.profile.string_table.append(string)
            return index

    def add_label_to_sample(self, sample: Sample, key: Any, value: Any) -> None:
        """Add a sample label to the given sample.

        The ``key`` and ``value`` arguments are both converted to strings.
        """
        _ = sample.label.add()
        _.key = self.get_string(str(key))
        _.str = self.get_string(str(value))

    def get_function(self, frame: Frame) -> int:
        """Get the function id from the given Austin frame."""
        key = (frame.function, frame.filename)
        try:
            return self._function_map[key]
        except KeyError:
            function = self.profile.function.add()
            function.id = len(self.profile.function)
            function.name = self.get_string(frame.function)
            function.filename = self.get_string(frame.filename)

            self._function_map[key] = function.id

            return function.id

    def get_location(self, frame: Frame) -> int:
        """Get the location id from the given Austin frame."""
        try:
            return self._location_map[frame]
        except KeyError:
            location = self.profile.location.add()
            location.id = len(self.profile.location)
            line = location.line.add()
            line.function_id = self.get_function(frame)
            line.line = frame.line

            self._location_map[frame] = location.id

            return location.id

    def add_samples(self, samples: List[AustinSample]) -> None:
        """Add a sample to the pprof generator."""
        # Create new pprof sample
        pprof_sample = self.profile.sample.add()

        # Add process and thread id labels
        self.add_label_to_sample(pprof_sample, THREAD_ID_LABEL, samples[0].thread)
        self.add_label_to_sample(pprof_sample, PROCESS_ID_LABEL, samples[0].pid)

        # Add metrics
        for sample in samples:
            pprof_sample.value.append(sample.metric.value)

        # Add locations. Top of the stack first.
        for frame in samples[0].frames[::-1]:
            pprof_sample.location_id.append(self.get_location(frame))

    def dump(self, stream: BinaryIO) -> None:
        """Dump the pprof protobuf message to the given binary stream."""
        stream.write(self.profile.SerializeToString())



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import dataclasses
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
import re
from typing import Any, Dict, Generator, Iterator, List, Optional, TextIO, Type, Union

from austin import AustinError

# ---- Custom types ----
ThreadName = str
ProcessId = int
MicroSeconds = int
KiloBytes = int


# ---- Exceptions ----


class InvalidFrame(AustinError):
    """Invalid frame.

    Thrown when attempting to parse a string that is supposed to represent a
    frame, but has the wrong structure.
    """

    pass


class InvalidSample(AustinError):
    """Invalid sample.

    Thrown when attempting to parse a string that is supposed to represent a
    sample, but has the wrong structure.
    """

    pass


# ---- Dataclasses ----


class Metadata(dict):
    """Austin Metadata."""

    def add(self, line: str) -> None:
        """Add a metadata line."""
        assert line.startswith("# ")
        key, _, value = line[2:].partition(":")
        self[key] = value.strip()


class MetricType(Enum):
    """Sample metric type."""

    TIME = 0
    MEMORY = 1

    @classmethod
    def from_mode(cls, mode: str) -> Optional["MetricType"]:
        """Convert metadata mode to metric type."""
        return {
            "cpu": MetricType.TIME,
            "wall": MetricType.TIME,
            "memory": MetricType.MEMORY,
            "full": None,
        }.get(mode)


@dataclass(frozen=True)
class Metric:
    """Austin metrics."""

    type: MetricType
    value: Union[MicroSeconds, KiloBytes] = 0

    def __add__(self, other: "Metric") -> "Metric":
        """Add metrics together (algebraically)."""
        assert self.type == other.type
        return Metric(
            type=self.type,
            value=self.value + other.value,
        )

    def __sub__(self, other: "Metric") -> "Metric":
        """Subtract metrics (algebraically)."""
        assert self.type == other.type
        return Metric(
            type=self.type,
            value=self.value - other.value,
        )

    def __gt__(self, other: "Metric") -> bool:
        """Strict comparison of metrics."""
        assert self.type == other.type
        return self.value > other.value

    def __ge__(self, other: "Metric") -> bool:
        """Comparison of metrics."""
        assert self.type == other.type
        return self.value >= other.value

    def copy(self) -> "Metric":
        """Make a copy of this object."""
        return dataclasses.replace(self)

    @staticmethod
    def parse(metrics: str, metric_type: Optional[MetricType] = None) -> List["Metric"]:
        """Parse the metrics from a sample.

        Returns a tuple containing the parsed metrics and the head of the
        sample for further processing.
        """
        try:
            ms = [int(_) for _ in metrics.split(",")]
            if len(ms) == 3:
                return [
                    Metric(MetricType.TIME, ms[0] if ms[1] == 0 else 0),
                    Metric(MetricType.TIME, ms[0]),
                    Metric(MetricType.MEMORY, ms[2] if ms[2] >= 0 else 0),
                    Metric(MetricType.MEMORY, -ms[2] if ms[2] < 0 else 0),
                ]
            elif len(ms) != 1:
                raise ValueError()

            assert metric_type is not None

            return [Metric(metric_type, ms[0])]

        except ValueError:
            raise InvalidSample(metrics)

    def __str__(self) -> str:
        """Stringify the metric."""
        return str(self.value)


@dataclass(frozen=True)
class Frame:
    """Python frame."""

    function: str
    filename: str
    line: int = 0

    @staticmethod
    def parse(frame: str) -> "Frame":
        """Parse the given string as a frame.

        A string representing a frame has the structure

            ``[frame] := <module>:<function>:<line number>``

        This static method attempts to parse the given string in order to
        identify the parts of the frame and returns an instance of the
        :class:`Frame` dataclass with the corresponding fields filled in.
        """
        if not frame:
            raise InvalidFrame(frame)

        try:
            module, function, line = frame.rsplit(":", maxsplit=3)
        except ValueError:
            raise InvalidFrame(frame)
        return Frame(function, module, int(line))

    def __str__(self) -> str:
        """Stringify the ``Frame`` object."""
        return f"{self.filename}:{self.function}:{self.line}"


@dataclass
class Sample:
    """Austin sample."""

    pid: ProcessId
    thread: ThreadName
    metric: Metric
    frames: List[Frame] = field(default_factory=list)

    _ALT_FORMAT_RE = re.compile(r";L([0-9]+)")

    @staticmethod
    def is_full(sample: str) -> bool:
        """Determine whether the sample has full metrics."""
        try:
            _, _, metrics = sample.rpartition(" ")
            return len(metrics.split(",")) == 3
        except (ValueError, IndexError):
            return False

    @staticmethod
    def parse(sample: str, metric_type: Optional[MetricType] = None) -> List["Sample"]:
        """Parse the given string as a frame.

        A string representing a sample has the structure

            ``P<pid>;T<tid>[;[frame]]* [metric][,[metric]]*``

        This static method attempts to parse the given string in order to
        identify the parts of the sample and returns an instance of the
        :class:`Sample` dataclass with the corresponding fields filled in.
        """
        if not sample:
            raise InvalidSample(sample)

        if sample[0] != "P":
            raise InvalidSample(f"No process ID in sample '{sample}'")

        head, _, metrics = sample.rpartition(" ")
        process, _, rest = head.partition(";")
        try:
            pid = int(process[1:])
        except ValueError:
            raise InvalidSample(f"Invalid process ID in sample '{sample}'")

        if rest[0] != "T":
            raise InvalidSample(f"No thread ID in sample '{sample}'")

        thread, _, frames = rest.partition(";")
        thread = thread[1:]

        if frames:
            if frames.rfind(";L"):
                frames = Sample._ALT_FORMAT_RE.sub(r":\1", frames)

        try:
            ms = Metric.parse(metrics, metric_type)
            return [
                Sample(
                    pid=int(pid),
                    thread=thread,
                    metric=metric,
                    frames=[Frame.parse(frame) for frame in frames.split(";")]
                    if frames
                    else [],
                )
                for metric in ms
            ]
        except ValueError as e:
            raise InvalidSample(f"Sample has invalid metric values: {sample}") from e
        except InvalidFrame as e:
            raise InvalidSample(f"Sample has invalid frames: {sample}") from e


@dataclass
class HierarchicalStats:
    """Base dataclass for representing hierarchical statistics.

    The statistics of a frame stack can be thought of as a rooted tree. Hence
    the hierarchy is established by the parent/child relation between the
    nodes in this tree. An instance of this class represents a node, and a
    leaf is given by those instances with an empty ``children`` attribute.

    The ``label`` attribute is used for indexing reasons and therefore should
    be of a hashable type.

    This class overrides the default ``add`` operator so that one can perform
    operations like ``stats1 + stats2``. Note, however, that instances of this
    class are not assumed to be immutable and indeed this operation will modify
    and return ``stats1`` with the outcome of the addition.
    """

    label: Any
    own: Metric
    total: Metric
    children: Dict[Any, "HierarchicalStats"] = field(default_factory=dict)

    def __lshift__(self, other: "HierarchicalStats") -> "HierarchicalStats":
        """Merge the RHS into the LHS."""
        if self.label != other.label:
            return self

        self.own += other.own
        self.total += other.total

        for frame, child in other.children.items():
            try:
                self.children[frame] << child
            except KeyError:
                self.children[frame] = child

        return self

    def get_child(self, label: Any) -> "HierarchicalStats":
        """Get a child from the children collection."""
        return self.children[label]

    def collapse(self, prefix: str = "") -> List[str]:
        """Collapse the hierarchical statistics."""
        if not self.children:
            return [f";{prefix}{self.label} {self.own}"]

        own = (
            []
            if self.own == Metric(MetricType.TIME)
            else [f";{prefix}{self.label} {self.own}"]
        )
        return own + [
            f";{prefix}{self.label}{rest}"
            for _, child in self.children.items()
            for rest in child.collapse()
        ]


@dataclass
class FrameStats(HierarchicalStats):
    """Frame statistics."""

    label: Frame
    height: int = 0
    children: Dict[Frame, "FrameStats"] = field(default_factory=dict)  # type: ignore[assignment]


class ThreadStats(HierarchicalStats):
    """Thread statistics."""

    label: ThreadName
    children: Dict[Frame, FrameStats] = field(default_factory=dict)  # type: ignore[assignment]

    def collapse(self, prefix: str = "T") -> List[str]:
        """Collapse the hierarchical statistics."""
        return super().collapse(prefix)


@dataclass
class ProcessStats:
    """Process statistics."""

    pid: ProcessId
    threads: Dict[ThreadName, ThreadStats] = field(default_factory=dict)

    def collapse(self) -> List[str]:
        """Collapse the hierarchical statistics."""
        return [
            f"P{self.pid}{rest}"
            for _, thread in self.threads.items()
            for rest in thread.collapse()
        ]

    def get_thread(self, thread_name: ThreadName) -> Optional[ThreadStats]:
        """Get thread statistics from this process by name.

        If the given thread name is not registered with this current process
        statistics, then ``None`` is returned.
        """
        return self.threads.get(thread_name)


class AustinStatsType(Enum):
    """Austin stats type."""

    WALL = "wall"
    CPU = "cpu"
    MEMORY_ALLOC = "memory_alloc"
    MEMORY_DEALLOC = "memory_dealloc"


class AustinFileReader:
    """Austin file reader.

    Conveniently read an Austin sample file by also parsing any header and
    footer metadata.
    """

    def __init__(self, file: str) -> None:
        self.file = file
        self.metadata = Metadata()
        self._stream: Optional[TextIO] = None
        self._stream_iter: Optional[Iterator] = None

    def _read_meta(self) -> None:
        assert self._stream_iter is not None

        for line in self._stream_iter:
            if not line.startswith("# ") or line == "\n":
                break
            self.metadata.add(line)

    def __enter__(self) -> "AustinFileReader":
        """Open the Austin file and read the metadata."""
        self._stream = open(self.file, "r")
        self._stream_iter = iter(self._stream)

        self._read_meta()

        return self

    def __iter__(self) -> Iterator:
        """Iterator over the samples in the Austin file."""

        def _() -> Generator[str, None, None]:
            assert self._stream_iter is not None

            for line in self._stream_iter:
                if line == "\n":
                    break
                yield line

            self._read_meta()

        return _()

    def __exit__(self, *args: Any) -> None:
        """Close the Austin file."""
        assert self._stream is not None

        self._stream.close()


@dataclass
class AustinStats:
    """Austin statistics.

    This class is used to collect all the statistics about own and total time
    and/or memory generated by a run of Austin. The :func:`update` method is
    used to pass a new :class:`Sample` so that the statistics can be updated
    accordingly.
    """

    stats_type: AustinStatsType
    processes: Dict[ProcessId, ProcessStats] = field(default_factory=dict)

    def dump(self, stream: TextIO) -> None:
        """Dump the statistics to the given text stream."""
        stream.write(f"# mode: {self.stats_type.value.partition('_')[0]}\n\n")
        for _, process in self.processes.items():
            samples = process.collapse()

            if all(sample.endswith(" 0 0") for sample in samples):
                samples = [sample[:-4] for sample in samples]

            for sample in samples:
                stream.write(sample + "\n")

    def get_process(self, pid: ProcessId) -> ProcessStats:
        """Get process statistics for the given PID."""
        return self.processes[pid]

    @classmethod
    def load(
        cls: Type["AustinStats"], stream: TextIO
    ) -> Dict[AustinStatsType, "AustinStats"]:
        """Load statistics from the given text stream."""
        meta = Metadata()
        for line in stream:
            if not line.startswith("# "):
                break
            meta.add(line)

        assert "mode" in meta
        metric_type = {
            "wall": MetricType.TIME,
            "cpu": MetricType.TIME,
            "memory": MetricType.MEMORY,
        }.get(meta["mode"])

        if metric_type is None:
            profiles = {
                AustinStatsType.CPU: AustinStats(AustinStatsType.CPU),
                AustinStatsType.WALL: AustinStats(AustinStatsType.WALL),
                AustinStatsType.MEMORY_ALLOC: AustinStats(AustinStatsType.MEMORY_ALLOC),
                AustinStatsType.MEMORY_DEALLOC: AustinStats(
                    AustinStatsType.MEMORY_DEALLOC
                ),
            }
        elif metric_type is MetricType.MEMORY:
            profiles = {
                AustinStatsType.MEMORY_ALLOC: AustinStats(AustinStatsType.MEMORY_ALLOC),
                AustinStatsType.MEMORY_DEALLOC: AustinStats(
                    AustinStatsType.MEMORY_DEALLOC
                ),
            }
        else:
            stats_type = (
                AustinStatsType.CPU if meta["mode"] == "cpu" else AustinStatsType.WALL
            )
            profiles = {stats_type: AustinStats(stats_type)}

        for line in stream:
            try:
                samples = Sample.parse(line, metric_type)
            except InvalidSample:
                continue

            if metric_type is None:
                cpu, wall, memory_alloc, memory_dealloc = samples

                profiles[AustinStatsType.WALL].update(wall)
                profiles[AustinStatsType.CPU].update(cpu)
                profiles[AustinStatsType.MEMORY_ALLOC].update(memory_alloc)
                profiles[AustinStatsType.MEMORY_DEALLOC].update(memory_dealloc)

            elif metric_type is MetricType.MEMORY:
                memory_alloc, memory_dealloc = samples

                profiles[AustinStatsType.MEMORY_ALLOC].update(memory_alloc)
                profiles[AustinStatsType.MEMORY_DEALLOC].update(memory_dealloc)

            else:
                profiles[stats_type].update(samples[0])

        return profiles

    def update(self, sample: Sample) -> None:
        """Update the statistics with a new sample.

        Normally, you would what to generate a new instance of :class:`Sample`
        by using :func:`Sample.parse` on a sample string passed by Austin to
        the sample callback.
        """
        zero = Metric(sample.metric.type, 0)
        pid = sample.pid
        thread_stats = ThreadStats(sample.thread, own=zero, total=sample.metric)

        # Convert the list of frames into a nested FrameStats instance
        stats: HierarchicalStats = thread_stats
        container = thread_stats.children
        for height, frame in enumerate(sample.frames):
            stats = FrameStats(
                label=frame, height=height, own=zero, total=sample.metric
            )
            container[frame] = stats
            container = stats.children
        stats.own = stats.total.copy()

        if pid not in self.processes:
            self.processes[pid] = ProcessStats(pid, {sample.thread: thread_stats})
            return

        process = self.processes[pid]
        if sample.thread not in process.threads:
            process.threads[sample.thread] = thread_stats
            return

        process.threads[sample.thread] << thread_stats



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import ArgumentParser
from difflib import SequenceMatcher
from io import StringIO
from typing import List, Set, TextIO, Tuple

from austin.format.compress import compress
from austin.stats import Frame
from austin.stats import InvalidSample
from austin.stats import Metric
from austin.stats import Sample

FoldedStack = List[Frame]


def _similarities(
    x: List[FoldedStack], y: List[FoldedStack]
) -> List[Tuple[Tuple[int, int], float]]:
    """O(n * log(n)), n = len(x) * len(y)."""

    def score(a: List[Frame], b: List[Frame]) -> float:
        """Score two folded stacks."""
        if not len(a) and not len(b):
            return 1
        return SequenceMatcher(a=a, b=b).ratio() - abs(len(a) - len(b)) / (
            len(a) + len(b)
        )

    return sorted(
        [((i, j), score(a, b)) for i, a in enumerate(x) for j, b in enumerate(y)],
        key=lambda x: x[1],
        reverse=True,
    )


def _match(
    x: List[FoldedStack], y: List[FoldedStack], threshold: float = 0.0
) -> Set[Tuple[int, int]]:
    """O(len(x) * len(y))."""
    ss = _similarities(x, y)
    mx, my = set(), set()
    matches = set()
    for (i, j), s in ss:
        if i in mx or j in my or s <= threshold:
            continue
        mx.add(i)
        my.add(j)
        matches.add((i, j))
    return matches


def diff(a: TextIO, b: TextIO) -> str:
    """Compare stacks and return a - b.

    The algorithm attempts to match stacks that look similar and returns
    only positive time differences (that is, stacks of b that took longer
    than those in a are not reported), plus any new stacks that are not in
    b.

    The return value is a string with collapsed stacks followed by the delta
    of the metrics on each line.
    """

    def compressed(source: TextIO) -> str:
        """Compress the source."""
        _ = StringIO()
        compress(source, _)
        return _.getvalue()

    def get_frames(text: str) -> List[Tuple[FoldedStack, Metric]]:
        """Get the folded stacks and metrics from a string of samples."""
        x = []
        for _ in text.splitlines(keepends=False):
            try:
                sample = Sample.parse(_)
                x.append((sample.frames, sample.metrics))
            except InvalidSample:
                continue
        return x

    fa = get_frames(compressed(a))
    fb = get_frames(compressed(b))
    ms = _match([frames for frames, _ in fa], [frames for frames, _ in fb])

    matched = set()
    stacks = []
    time_only = True

    # Matched stacks
    for i, j in ms:
        matched.add(i)
        delta = fa[i][1] - fb[j][1]
        if delta >= 0:
            stacks.append((fa[i][0], delta))
        if delta.memory_alloc or delta.memory_dealloc:
            time_only = False

    # New stacks
    for i in [_ for _ in range(len(fa)) if _ not in matched]:
        f, m = fa[i]
        stacks.append((f, m))

    return "\n".join(
        [
            ";".join([str(_) for _ in f]) + " " + str(m.time if time_only else m)
            for f, m in stacks
        ]
    )


def main() -> None:
    """Diff tool for Austin samples."""
    argp = ArgumentParser(
        prog="austin-diff",
        description=("Compute the diff between two austin frame stack sample files"),
    )

    argp.add_argument("a", type=str, help="The minuend collapsed stacks")
    argp.add_argument("b", type=str, help="The subtrahend collapsed stacks")
    argp.add_argument("-o", "--output", type=str, help="The output file")
    argp.add_argument("-V", "--version", action="version", version="0.1.0")

    args = argp.parse_args()

    with open(args.a) as a, open(args.b) as b:
        result = diff(a, b)
        if args.output is not None:
            with open(args.output, "w") as fout:
                fout.write(result)
        else:
            print(result)


if __name__ == "__main__":
    main()



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import ArgumentParser

from austin.format.pprof import PProf
from austin.stats import AustinFileReader
from austin.stats import InvalidSample
from austin.stats import MetricType
from austin.stats import Sample


def main() -> None:
    """austin2pprof entry point."""
    arg_parser = ArgumentParser(
        prog="austin2pprof",
        description=(
            "Convert Austin generated profiles to the pprof protobuf format. "
            "See https://github.com/google/pprof for more details."
        ),
    )

    arg_parser.add_argument(
        "input",
        type=str,
        help="The input file containing Austin samples.",
    )
    arg_parser.add_argument(
        "output", type=str, help="The name of the output pprof file."
    )

    arg_parser.add_argument("-V", "--version", action="version", version="0.1.0")

    args = arg_parser.parse_args()

    try:
        with AustinFileReader(args.input) as fin:
            mode = fin.metadata["mode"]
            pprof = PProf(mode=mode)

            # Read samples
            for line in fin:
                try:
                    pprof.add_samples(Sample.parse(line, MetricType.from_mode(mode)))
                except InvalidSample:
                    continue

    except FileNotFoundError:
        print(f"No such input file: {args.input}")
        exit(1)

    with open(args.output, "wb") as fout:
        pprof.dump(fout)


if __name__ == "__main__":
    main()



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from threading import Thread
from typing import Any, List, Optional

from austin import AustinError
from austin.simple import SimpleAustin


class ThreadedAustin(SimpleAustin):
    """Thread-based implementation of Austin.

    Implements a ``threading`` API for Austin so that it can be used alongside
    other threads.

    The following example shows how to make a simple threaded echo
    implementation of Austin that behaves exactly just like Austin.

    Example::

        class EchoThreadedAustin(ThreadedAustin):
            def on_ready(self, process, child_process, command_line):
                print(f"Austin PID: {process.pid}")
                print(f"Python PID: {child_process.pid}")
                print(f"Command Line: {command_line}")

            def on_sample_received(self, line):
                print(line)

            def on_terminate(self, data):
                print(data)

        try:
            austin = EchoThreadedAustin()
            austin.start(["-i", "10000", "python3", "myscript.py"])
            austin.join()
        except KeyboardInterrupt:
            pass
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._thread: Optional[Thread] = None
        self._exc: Optional[Exception] = None

    def start(self, args: List[str] = None) -> None:
        """Start the Austin thread."""

        def _thread_bootstrap() -> None:
            try:
                SimpleAustin.start(self, args)
            except Exception as e:
                self._exc = e

        self._thread = Thread(target=_thread_bootstrap)
        self._thread.start()

    def get_thread(self) -> Optional[Thread]:
        """Get the underlying :class:`threading.Thread` instance.

        As this leaks a bit of the implementation, interaction with the
        returned thread object should be done with care.

        **Note**
            This is an extension of the base Austin abstract class.
        """
        return self._thread

    def join(self, timeout: float = None) -> None:
        """Join the thread.

        This is the same as calling :func:`join` on the underlying thread
        object.

        **Note**
            This is an extension of the base Austin abstract class.
        """
        if self._thread is None:
            raise AustinError("Austin thread has not been started yet")
        self._thread.join(timeout)
        if self._exc:
            raise self._exc



from enum import Enum


class Mode(Enum):
    """Austin profiling mode."""

    CPU = 0
    WALL = 1
    MEMORY = 2
    FULL = 3

    @classmethod
    def from_metadata(cls, mode: str) -> "Mode":
        """Get mode from metadata information."""
        return {
            "cpu": Mode.CPU,
            "wall": Mode.WALL,
            "memory": Mode.MEMORY,
            "full": Mode.FULL,
        }[mode]



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, TextIO

from austin.stats import AustinFileReader


def compress(source: AustinFileReader, dest: TextIO, counts: bool = False) -> None:
    """Compress the source stream.

    Aggregates the metrics on equal collapsed stacks. If ``counts`` is ``True``
    then time samples are counted by occurrence rather than by sampling
    duration.
    """
    stats: Dict[str, int] = {}

    for line in source:
        head, _, metric = line.rpartition(" ")
        if "," in metric:
            raise RuntimeError("Cannot compress samples with full metrics.")

        value = 1 if counts else int(metric)
        if metric:
            stats[head] = stats.get(head, 0) + value

    dest.writelines([f"# {k}: {v}\n" for k, v in source.metadata.items()])
    dest.write("\n")
    dest.writelines([head + " " + str(metric) + "\n" for head, metric in stats.items()])


def main() -> None:
    """austin-compress entry point."""
    from argparse import ArgumentParser

    arg_parser = ArgumentParser(
        prog="austin-compress",
        description=(
            "Compress a raw Austin sample file by aggregating the collected samples."
        ),
    )

    arg_parser.add_argument(
        "input",
        type=str,
        help="The input file containing Austin samples.",
    )
    arg_parser.add_argument(
        "output",
        type=str,
        help="The output file; defaults to the input file for in-place compression.",
    )

    arg_parser.add_argument(
        "-c",
        "--counts",
        action="store_true",
        default=False,
        help="Use sample counts instead of measured sampling durations.",
    )

    arg_parser.add_argument("-V", "--version", action="version", version="0.2.0")

    args = arg_parser.parse_args()

    try:
        with AustinFileReader(args.input) as fin, open(
            args.output or args.input, "w"
        ) as fout:
            compress(fin, fout, args.counts)
    except FileNotFoundError:
        print(f"No such input file: {args.input}")
        exit(1)


if __name__ == "__main__":
    main()



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os.path
from typing import Any, Dict

import toml


class AustinConfigurationError(Exception):
    """Austin configuration error."""

    pass


class AustinConfiguration:
    """Austin configuration borg."""

    RC = os.path.join(os.path.expanduser("~"), ".austinrc")

    __borg__: Dict[str, Any] = {}

    def __init__(self) -> None:
        self.__dict__ = self.__borg__
        self.reload()

    def reload(self) -> None:
        """Reload the configuration from the run-control file."""
        try:
            with open(self.RC) as fin:
                self._data = toml.load(fin)
        except FileNotFoundError:
            self._data = {}

    def save(self) -> None:
        """Save the current configuration to file."""
        try:
            with open(self.RC, "w") as fout:
                toml.dump(self._data, fout)
        except Exception as e:
            raise AustinConfigurationError("Unable to save Austin configuration") from e

    @property
    def binary(self) -> str:
        """Get the Austin binary path."""
        return self._data.get("binary", None)



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
import json
from typing import Dict, List, Optional, TextIO, Union

from austin.format import Mode
from austin.stats import AustinFileReader
from austin.stats import Frame
from austin.stats import InvalidSample
from austin.stats import MetricType
from austin.stats import Sample

SpeedscopeJson = Dict
SpeedscopeWeight = int
ProfileName = str


class Units(Enum):
    """Metric units."""

    MICROSECONDS = "μs"
    BYTES = "bytes"


@dataclass(frozen=True)
class SpeedscopeFrame:
    """Speedscope Frame object."""

    name: str
    file: str
    line: int


@dataclass
class SpeedscopeProfile:
    """Speedscope Profile object."""

    name: ProfileName
    unit: str
    startValue: int = 0
    endValue: int = 0
    samples: List[List[int]] = field(default_factory=list)
    weights: List[SpeedscopeWeight] = field(default_factory=list)
    type: str = "sampled"

    def add_sample(self, stack: List[int], weight: SpeedscopeWeight) -> None:
        """Add a sample to the profile."""
        self.samples.append(stack)
        self.weights.append(weight)
        self.endValue += weight


class Speedscope:
    """Speedscope JSON generator."""

    def __init__(
        self, name: str, mode: Union[Mode, str], indent: Optional[int] = None
    ) -> None:
        self.name = name
        self.indent = indent
        self.mode = mode
        self.mode = Mode.from_metadata(mode) if isinstance(mode, str) else mode

        self.profiles: List[SpeedscopeProfile] = []
        self.profile_map: Dict[int, Dict[str, Dict[str, SpeedscopeProfile]]] = {}

        self.frames: List[dict] = []
        self.frame_map: Dict[Frame, int] = {}

    def get_frame(self, frame: Frame) -> int:
        """Get the index of an observed frame."""
        if frame in self.frame_map:
            return self.frame_map[frame]

        index = len(self.frames)
        self.frame_map[frame] = index
        self.frames.append(
            asdict(SpeedscopeFrame(frame.function, frame.filename, frame.line))
        )

        return index

    def get_profile(self, pid: int, thread: str, metric: str) -> SpeedscopeProfile:
        """Get the profile for the given pid, thread and profile metric."""
        prefix = {
            "cpu": "CPU time",
            "wall": "Wall time",
            "m+": "Memory allocation",
            "m-": "Memory deallocation",
        }[metric]
        units = Units.BYTES if metric[0] == "m" else Units.MICROSECONDS
        profiles = self.profile_map.setdefault(pid, {}).setdefault(thread, {})
        if metric in profiles:
            return profiles[metric]

        self.profiles.append(
            SpeedscopeProfile(
                name=f"{prefix} profile for {pid}:{thread}",
                unit=units.value,
            )
        )
        return profiles.setdefault(
            metric,
            self.profiles[-1],
        )

    def add_samples(self, samples: List[Sample]) -> None:
        """Add a sample to the generator."""
        if self.mode == Mode.CPU:
            _ = zip(("cpu",), samples)
        elif self.mode == Mode.WALL:
            _ = zip(("cpu",), samples)
        elif self.mode == Mode.MEMORY:
            _ = zip(("m+", "m-"), samples)
        elif self.mode == Mode.FULL:
            _ = zip(("cpu", "wall", "m+", "m-"), samples)

        for prefix, sample in _:
            if not sample.frames or sample.metric.value == 0:
                continue

            self.get_profile(sample.pid, sample.thread, prefix).add_sample(
                [self.get_frame(frame) for frame in sample.frames], sample.metric.value
            )

    def asdict(self) -> SpeedscopeJson:
        """Return the JSON as a Python dictionary."""
        return {
            "$schema": "https://www.speedscope.app/file-format-schema.json",
            "shared": {"frames": self.frames},
            "profiles": sorted(
                [asdict(profile) for profile in self.profiles],
                key=lambda p: p["name"].rsplit(maxsplit=1)[-1],
            ),
            "name": self.name,
            "exporter": "Austin2Speedscope Converter 0.2.0",
        }

    def dump(self, stream: TextIO) -> None:
        """Dump the pprof protobuf message to the given binary stream."""
        json.dump(
            self.asdict(),
            stream,
            indent=self.indent,
        )


def main() -> None:
    """austin2speedscope entry point."""
    from argparse import ArgumentParser
    import os

    arg_parser = ArgumentParser(
        prog="austin2speedscope",
        description=(
            "Convert Austin generated profiles to the Speedscope JSON format "
            "accepted by https://speedscope.app. The output will contain a profile "
            "for each thread and metric included in the input file."
        ),
    )

    arg_parser.add_argument(
        "input",
        type=str,
        help="The input file containing Austin samples in normal format.",
    )
    arg_parser.add_argument(
        "output", type=str, help="The name of the output Speedscope JSON file."
    )
    arg_parser.add_argument(
        "--indent", type=int, help="Give a non-null value to prettify the JSON output."
    )

    arg_parser.add_argument("-V", "--version", action="version", version="0.1.0")

    args = arg_parser.parse_args()

    try:
        with AustinFileReader(args.input) as fin:
            mode = fin.metadata["mode"]
            speedscope = Speedscope(os.path.basename(args.input), mode, args.indent)
            for line in fin:
                try:
                    speedscope.add_samples(
                        Sample.parse(line, MetricType.from_mode(mode))
                    )
                except InvalidSample:
                    continue

    except FileNotFoundError:
        print(f"No such input file: {args.input}")
        exit(1)

    with open(args.output, "w") as fout:
        speedscope.dump(fout)


if __name__ == "__main__":
    main()



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import ArgumentParser
from argparse import Namespace
from argparse import REMAINDER
from typing import Any, Callable, List, NoReturn

from austin import AustinError


class AustinCommandLineError(AustinError):
    """Invalid Austin command line."""

    pass


class AustinArgumentParser(ArgumentParser):
    """Austin Command Line parser.

    This command line parser is based on :class:`argparse.ArgumentParser` and
    provides a minimal implementation for parsing the standard Austin command
    line. The bool arguments of the constructor are used to specify whether
    the corresponding Austin option should be parsed or not. For example, if
    your application doesn't need the possiblity of switching to the
    alternative format, you can exclude this option with ``alt_format=False``.

    Note that al least one between ``pid`` and ``command`` is required, but
    they cannot be used together when invoking Austin.
    """

    def __init__(
        self,
        name: str = "austin",
        alt_format: bool = True,
        children: bool = True,
        exclude_empty: bool = True,
        exposure: bool = True,
        full: bool = True,
        interval: bool = True,
        memory: bool = True,
        pid: bool = True,
        sleepless: bool = True,
        timeout: bool = True,
        command: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(prog=name, **kwargs)

        def time(units: str) -> Callable[[str], int]:
            """Parse time argument with units."""
            base = int({"us": 1, "ms": 1e3, "s": 1e6}[units])

            def parser(arg: str) -> int:
                if arg.endswith("us"):
                    return int(arg[:-2]) // base
                if arg.endswith("ms"):
                    return int(arg[:-2]) * 1000 // base
                if arg.endswith("s"):
                    return int(arg[:-1]) * 1000000 // base
                return int(arg)

            return parser

        if not (pid and command):
            raise AustinCommandLineError(
                "Austin command line parser must have at least one between pid "
                "and command."
            )

        if alt_format:
            self.add_argument(
                "-a",
                "--alt-format",
                help="Alternative collapsed stack sample format.",
                action="store_true",
            )

        if children:
            self.add_argument(
                "-C",
                "--children",
                help="Attach to child processes.",
                action="store_true",
            )

        if exclude_empty:
            self.add_argument(
                "-e",
                "--exclude-empty",
                help="Do not output samples of threads with no frame stacks.",
                action="store_true",
            )

        if exposure:
            self.add_argument(
                "-x",
                "--exposure",
                help="Sample for the given number of seconds only.",
                type=time("s"),
                default=None,
            )

        if full:
            self.add_argument(
                "-f",
                "--full",
                help="Produce the full set of metrics (time +mem -mem).",
                action="store_true",
            )

        if interval:
            self.add_argument(
                "-i",
                "--interval",
                help="Sampling interval (default is 100 μs).",
                type=time("us"),
            )

        if memory:
            self.add_argument(
                "-m", "--memory", help="Profile memory usage.", action="store_true"
            )

        if pid:
            self.add_argument(
                "-p",
                "--pid",
                help="The the ID of the process to which Austin should attach.",
                type=int,
            )

        if sleepless:
            self.add_argument(
                "-s", "--sleepless", help="Suppress idle samples.", action="store_true"
            )

        if timeout:
            self.add_argument(
                "-t",
                "--timeout",
                help="Approximate start up wait time. Increase on slow machines "
                "(default is 100 ms).",
                type=time("ms"),
            )

        if command:
            self.add_argument(
                "command",
                type=str,
                nargs=REMAINDER,
                help="The command to execute if no PID is provided, followed by "
                "its arguments.",
            )

    def parse_args(  # type: ignore[override]
        self, args: List[str] = None, namespace: Namespace = None
    ) -> Namespace:
        """Parse the list of arguments.

        Return a :class:`argparse.Namespace` with the parsed result. If no PID
        nor a command are passed, an instance of the
        :class:`AustinCommandLineError` exception is thrown.
        """
        parsed_austin_args, unparsed = super().parse_known_args(args, namespace)
        if unparsed:
            raise AustinCommandLineError(
                f"Some arguments were left unparsed: {unparsed}"
            )

        if not parsed_austin_args.pid and not parsed_austin_args.command:
            raise AustinCommandLineError("No PID or command given.")

        return parsed_austin_args

    def exit(self, status: int = 0, message: str = None) -> NoReturn:
        """Raise exception on error."""
        raise AustinCommandLineError(message, status)

    @staticmethod
    def to_list(args: Namespace) -> List[str]:
        """Convert a :class:`argparse.Namespace` to a list of arguments.

        This is the opposite of the parsing of the command line. This static
        method is intended to filter and reconstruct the command line arguments
        that need to be passed to lower level APIs to start the actual Austin
        process.
        """
        arg_list = []
        if getattr(args, "alt_format", None):
            arg_list.append("-a")
        if getattr(args, "children", None):
            arg_list.append("-C")
        if getattr(args, "exclude_empty", None):
            arg_list.append("-e")
        if getattr(args, "full", None):
            arg_list.append("-f")
        if getattr(args, "interval", None):
            arg_list += ["-i", str(args.interval)]
        if getattr(args, "memory", None):
            arg_list.append("-m")
        if getattr(args, "pid", None):
            arg_list += ["-p", str(args.pid)]
        if getattr(args, "sleepless", None):
            arg_list.append("-s")
        if getattr(args, "timeout", None):
            arg_list += ["-t", str(args.timeout)]
        if getattr(args, "command", None):
            arg_list += args.command

        return arg_list



# This file is part of "austin-python" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-python is a Python wrapper around Austin, the CPython frame stack
# sampler.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import sys
from typing import Dict, List, Optional

from austin import AustinError
from austin import BaseAustin
from austin.cli import AustinArgumentParser


class AsyncAustin(BaseAustin):
    """Asynchronous implementation of Austin.

    Implements an ``asyncio`` API for Austin so that it can be used alongside
    other asynchronous tasks.

    The following example shows how to make a simple asynchronous echo
    implementation of Austin that behaves exactly just like Austin.

    Example::

        class EchoAsyncAustin(AsyncAustin):
            def on_ready(self, process, child_process, command_line):
                print(f"Austin PID: {process.pid}")
                print(f"Python PID: {child_process.pid}")
                print(f"Command Line: {command_line}")

            def on_sample_received(self, line):
                print(line)

            def on_terminate(self, data):
                print(data)

        if sys.platform == "win32":
            asyncio.set_event_loop(asyncio.ProactorEventLoop())

        try:
            austin = EchoAsyncAustin()
            asyncio.get_event_loop().run_until_complete(
                austin.start(["-i", "10000", "python3", "myscript.py"])
            )
        except (KeyboardInterrupt, asyncio.CancelledError):
            pass
    """

    async def _read_stderr(self) -> Optional[str]:
        assert self.proc.stderr is not None

        try:
            return (
                (await asyncio.wait_for(self.proc.stderr.read(), 0.1)).decode().rstrip()
            )
        except asyncio.TimeoutError:
            return None

    async def _read_meta(self) -> Dict[str, str]:
        assert self.proc.stdout is not None

        meta = {}

        while True:
            line = (await self.proc.stdout.readline()).decode().rstrip()
            if not (line and line.startswith("# ")):
                break
            key, _, value = line[2:].partition(": ")
            meta[key] = value

        self._meta.update(meta)
        return meta

    async def start(self, args: List[str] = None) -> None:
        """Create the start coroutine.

        Use with the ``asyncio`` event loop.
        """
        try:
            _args = list(args or sys.argv[1:])
            _args.insert(0, "-P")
            self.proc = await asyncio.create_subprocess_exec(
                self.binary_path,
                *_args,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
        except FileNotFoundError:
            raise AustinError("Austin executable not found.")

        if not self.proc.stdout:
            raise AustinError("Standard output stream is unexpectedly missing")
        if not self.proc.stderr:
            raise AustinError("Standard error stream is unexpectedly missing")

        self._running = True

        try:
            if not await self._read_meta():
                raise AustinError("Austin did not start properly")

            self.check_version()

            self._ready_callback(
                *self._get_process_info(
                    AustinArgumentParser().parse_args(args), self.proc.pid
                )
            )

            # Start readline loop
            while self._running:
                data = (await self.proc.stdout.readline()).rstrip()
                if not data:
                    break

                self.submit_sample(data)

            self._terminate_callback(await self._read_meta())
            self.check_exit(await self.proc.wait(), await self._read_stderr())

        except Exception:
            self.proc.terminate()
            await self.proc.wait()
            raise

        finally:
            self._running = False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['austin', 'austin.format', 'austin.format.pprof', 'austin.tools']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses',
 'protobuf>=3.12.2,<4.0.0',
 'psutil>=5.7.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['austin-compress = austin.format.compress:main',
                     'austin-diff = austin.tools.diff:main',
                     'austin2pprof = austin.format.pprof.__main__:main',
                     'austin2speedscope = austin.format.speedscope:main']}

setup_kwargs = {
    'name': 'austin-python',
    'version': '1.0.0a3',
    'description': 'Python wrapper for Austin, the frame stack sampler for CPython',
    'long_description': '<p align="center">\n  <br/>\n  <img src="docs/source/images/logo.png"\n       alt="Austin"\n       height="256px" />\n  <br/>\n</p>\n\n<h3 align="center">Python wrapper for Austin</h3>\n\n<p align="center">\n  <a href="https://github.com/P403n1x87/austin-python/actions?workflow=Tests">\n    <img src="https://github.com/P403n1x87/austin-python/workflows/Tests/badge.svg"\n         alt="GitHub Actions: Tests">\n  </a>\n  <a href="https://travis-ci.org/P403n1x87/austin-python">\n    <img src="https://travis-ci.org/P403n1x87/austin-python.svg?branch=main"\n         alt="Travis CI">\n  </a>\n  <a href="https://codecov.io/gh/P403n1x87/austin-python">\n    <img src="https://codecov.io/gh/P403n1x87/austin-python/branch/main/graph/badge.svg"\n         alt="Codecov">\n  </a>\n  <a href="https://austin-python.readthedocs.io/">\n    <img src="https://readthedocs.org/projects/austin-python/badge/"\n         alt="Documentation">\n  </a>\n  <br/>\n  <a href="https://pypi.org/project/austin-python/">\n    <img src="https://img.shields.io/pypi/v/austin-python.svg"\n         alt="PyPI">\n  </a>\n  <a href="https://pepy.tech/project/austin-python">\n    <img src="https://static.pepy.tech/personalized-badge/austin-python?period=total&units=international_system&left_color=grey&right_color=blue&left_text=downloads"\n         alt="Downloads" />\n  <a/>\n  <br/>\n  <a href="https://github.com/P403n1x87/austin-python/blob/main/LICENSE.md">\n    <img src="https://img.shields.io/badge/license-GPLv3-ff69b4.svg"\n         alt="LICENSE">\n  </a>\n</p>\n\n<p align="center">\n  <a href="#synopsis"><b>Synopsis</b></a>&nbsp;&bull;\n  <a href="#installation"><b>Installation</b></a>&nbsp;&bull;\n  <a href="#usage"><b>Usage</b></a>&nbsp;&bull;\n  <a href="#compatibility"><b>Compatibility</b></a>&nbsp;&bull;\n  <a href="#documentation"><b>Documentation</b></a>&nbsp;&bull;\n  <a href="#contribute"><b>Contribute</b></a>\n</p>\n\n<p align="center">\n  <a href="https://www.buymeacoffee.com/Q9C1Hnm28" target="_blank">\n    <img src="https://www.buymeacoffee.com/assets/img/custom_images/orange_img.png" alt="Buy Me A Coffee" />\n  </a>\n</p>\n\n\n# Synopsis\n\nThe `austin-python` package is a Python wrapper around the [Austin] binary that\nprovides convenience classes to quickly develop your statistical profiling\ntools. Whether your code is thread-based or asynchronous, `austin-python` has\nyou covered. This is, for instance, how you would turn Austin into a Python\napplication:\n\n~~~ python\nfrom austin.aio import AsyncAustin\n\n\n# Make your sub-class of AsyncAustin\nclass EchoAsyncAustin(AsyncAustin):\n    def on_ready(self, process, child_process, command_line):\n        print(f"Austin PID: {process.pid}")\n        print(f"Python PID: {child_process.pid}")\n        print(f"Command Line: {command_line}")\n\n    def on_sample_received(self, line):\n        print(line)\n\n    def on_terminate(self, data):\n        print(data)\n\n\n# Use the Proactor event loop on Windows\nif sys.platform == "win32":\n    asyncio.set_event_loop(asyncio.ProactorEventLoop())\n\ntry:\n    # Start the Austin application with some command line arguments\n    austin = EchoAsyncAustin()\n    asyncio.get_event_loop().run_until_complete(\n        austin.start(["-i", "10000", "python3", "myscript.py"])\n    )\nexcept (KeyboardInterrupt, asyncio.CancelledError):\n    pass\n~~~\n\nThe `austin-python` package is at the heart of the [Austin\nTUI](https://github.com/P403n1x87/austin-tui) and the [Austin\nWeb](https://github.com/P403n1x87/austin-web) Python applications. Go check them\nout if you are looking for full-fledged usage examples.\n\nIncluded with the package come two applications for the conversion of Austin\ncollected output, which is in the form of [collapsed\nstacks](https://github.com/brendangregg/FlameGraph), to either the\n[Speedscope](https://speedscope.app/) JSON format or the [Google pprof\nformat](https://github.com/google/pprof). Note, however, that the Speedscope web\napplication supports Austin native format directly.\n\n\n# Installation\n\nThis package can be installed from PyPI with\n\n~~~ bash\npip install --user austin-python --upgrade\n~~~\n\nPlease note that `austin-python` requires the [Austin] binary. The default\nlookup locations are, in order,\n\n- current working directory;\n- the `AUSTINPATH` environment variable which gives the path to the folder that\n  contains the Austin binary;\n- the `.austinrc` TOML configuration file in the user\'s home folder, e.g.\n  `~/.austinrc` on Linux (see below for a sample `.austinrc` file);\n- the `PATH` environment variable.\n\nA sample `.austinrc` file would look like so\n\n~~~ toml\nbinary = "/path/to/austin"\n~~~\n\n\n# Usage\n\nA simple example of an echo application was shown above. Other examples using,\ne.g., threads, can be found in the official documentation. You can also browse\nthrough the code of the [Austin TUI](https://github.com/P403n1x87/austin-tui)\nand the [Austin Web](https://github.com/P403n1x87/austin-web) Python\napplications to see how they leverage `austin-python`.\n\n## Format conversion\n\nAs it was mentioned before, this package also comes with two scripts for format\nconversion, namely `austin2speedscope` and `austin2pprof`. They both take two\nmandatory arguments, that is, the input and output file. For example, to convert\nthe Austin profile data file `myscript.aprof` to the Google pprof data file\n`myscript.pprof`, you can run\n\n~~~ bash\naustin2pprof myscript.aprof myscript.pprof\n~~~\n\nThe package also provide the `austin-compress` utility to compress the Austin\nraw samples by aggregation.\n\n# Compatibility\n\nThe `austin-python` package is tested on Linux, macOS and Windows with Python\n3.6-3.9.\n\n\n# Documentation\n\nThe official documentation is hosted on readthedocs.io at\n[austin-python.readthedocs.io](https://austin-python.readthedocs.io/).\n\n\n# Contribute\n\nIf you want to help with the development, then have a look at the open issues\nand have a look at the [contributing guidelines](CONTRIBUTING.md) before you\nopen a pull request.\n\nYou can also contribute to the development by either [becoming a\nPatron](https://www.patreon.com/bePatron?u=19221563) on Patreon, by [buying me a\ncoffee](https://www.buymeacoffee.com/Q9C1Hnm28) on BMC or by chipping in a few\npennies on [PayPal.Me](https://www.paypal.me/gtornetta/1).\n\n<p align="center">\n  <a href="https://www.buymeacoffee.com/Q9C1Hnm28" target="_blank">\n    <img src="https://www.buymeacoffee.com/assets/img/custom_images/orange_img.png"\n         alt="Buy Me A Coffee" />\n  </a>\n</p>\n\n\n[Austin]: https://github.com/p403n1x87/austin\n',
    'author': 'Gabriele N. Tornetta',
    'author_email': 'phoenix1987@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/P403n1x87/austin-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
